import numpy as np
import matplotlib.pylab as plt
import matplotlib.animation as animation
import casadi as ca
from rk4 import rk4
from plot_helper import animate_pendulum
plt.close("all")

# parameters
T = 30                  # time length
N = 150                 # number of discrete time points
h = T/N                 # one discrete time step
nx = 2                  # state dimension
nu = 1                  # control dimension

x0bar = [np.pi, 0]      # initial state
xNbar = [0, 0]          # target state
umax = .13              # maximum control torque

# set up dynamics
# casadi variables independent of the later declared Opti() instance
x = ca.SX.sym('x', nx)      # general state
u = ca.SX.sym('u', nu)      # general control
# declare names for convenience
theta = x[0]
omega = x[1]
# ode of dynamics
xdot = ca.vertcat(omega, ca.sin(theta) + u)
# Declare casadi function for ode
f = ca.Function('f', [x, u], [xdot])
# Declare casadi function for integration
F = ca.Function('F', [x, u], [rk4(f, x, u, h)])

# start defining the NLP
opti = ca.Opti()

# decision variables
X = opti.variable(nx, N+1)      # X[:,k] is state at discrete time k
U = opti.variable(nu, N)        # U[:,k] is control at discrete time k

# objective
opti.minimize(.5 * ca.sum2( U**2 ))

# initial position constraint
opti.subject_to(X[:,0] == x0bar)
# dynamics
for k in range(N):
    opti.subject_to(X[:, k+1] == F(X[:,k], U[:, k]))
# target constraint
opti.subject_to(X[:,-1] == xNbar)
# control constraint
opti.subject_to(opti.bounded(-umax, U, umax))

# initialize all states at x0bar
opti.set_initial(X, ca.repmat(x0bar, 1, N+1) )

# solve with IPOPT
opti.solver('ipopt')
sol = opti.solve()

# read solution
X_opt = sol.value(X)
U_opt = sol.value(U)
Theta_opt = X_opt[0, :]
Omega_opt = X_opt[1, :]

# plot state trajectory
plt.figure()
plt.plot(Theta_opt, '.-', label='theta')
plt.plot(Omega_opt, '.-', label='omega')
plt.legend()

# plot control trajectory
plt.figure()
plt.step(range(N), U_opt, label='u', where='post')
plt.legend()

# animate swing up
ani = animate_pendulum(Theta_opt, dt=.03)
plt.show()

#%%
# simulate trajectory without any control
# (this was no task on the sheet)

Nsim = 90

X_all = np.zeros((nx, Nsim+1))
X_all[0,0] = 3/4 * np.pi
for i in range(Nsim):
    xnext = F(X_all[:, i], 0)
    # xnext = F2(X_all[:, i], 0, h)
    X_all[:,i+1] = xnext.full().flatten()
    
    
# Visualize 
theta = 2.2
Theta = np.linspace(0, theta, 30)
fac = .2

plt.figure(figsize=(8,3))
plt.subplot(121)
plt.plot([0, 0], [0, 1], 'k--')
plt.plot(-fac * np.sin(Theta), fac * np.cos(Theta), 'k')
plt.text(-.14, .01, r'$\theta$')
plt.plot([0, -np.sin(theta)], [0, np.cos(theta)], '.-')

a = 1.1
plt.plot([-a, a, a, -a, -a], [a, a, -a, -a, a], 'k', lw=.5)

plt.axis('equal')
plt.axis('off')
plt.xlim((-1.1, 1.1))
plt.ylim((-1.1, 1.1))  
    
# plt.figure(figsize=(5,3))
plt.subplot(122)

plt.plot(X_all[0,:], '.-')
plt.xlabel(r'diskrete Zeit $k$')
plt.ylabel(r'Pendelposition $\theta$')

# plt.tight_layout()
# plt.savefig('pendulum.pdf')
